/*
 * ZipSnap --- Directory Backup Tool
 * v 0.2 (2006-03-19)
 * Copyright 2006 Zach Scrivena
 * Email: zachscrivena@gmail.com
 *
 * ZipSnap is a simple command-line utility that creates compressed
 * point-in-time snapshots of directories.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

import java.util.*;
import java.util.zip.*;
import java.io.*;
import java.text.*;

/**
 * Performs all the functions provided by the ZipSnap utility.
 *
 */
public class ZipSnap {

	// constants
	private static final String progTitle = "ZipSnap 0.2   Copyright 2006 Zach Scrivena   2006-03-19";
	private static final int BUFFER_SIZE = 1048576;		// 1 Mb
	private static final int DEFLATED_MIN = 256;		// 256 bytes
	private static final String NAME_SEPARATOR = "/";	// system-neutral name separator used in file names in catalog

	// enums and default program modes
	private static enum Overwrite {Ask, Always, Never};
	private static Overwrite overwriteMode = Overwrite.Ask;
	private static enum DeleteExisting {Ask, Always, Never};
	private static DeleteExisting deleteExistingMode = DeleteExisting.Ask;
	private static enum Checksum {Use, Ignore};
	private static Checksum checksumMode = Checksum.Ignore;
	private static String argCatalog = "";
	private static String argSnapshotNumber = "";
	private static String argVolumeNumber = "";
	private static String argTimeAllowance = "";

	// ZipSnap archive variables
	private static ArrayList<SnapshotEntry> snapshots = null;
	private static ArrayList<VolumeEntry> volumes = null;
	private static ArrayList<FileEntry> files = null;

	// variables
	private static byte byteBuffer[] = new byte[BUFFER_SIZE];
	private static String catalogName = null;
	private static File catalogFile = null;
	private static String directoryName = null;
	private static File directoryFile = null;
	private static String workName = null;
	private static File workFile = null;


	/**
	 * Prints usage documentation.
	 */
    private static void printUsage() {
		// RULER
		// 00000000011111111112222222222333333333344444444445555555555666666666677777777778
		// 12345678901234567890123456789012345678901234567890123456789012345678901234567890
		p(""+
		"\nZipSnap creates compressed point-in-time snapshots of directories."+
		"\n"+
		"\nUSAGE:   java -jar ZipSnap.jar  [command]  <switches>  [catalog]"+
		"\n"+
		"\n  [Commands]:"+
		"\n     a    Add a snapshot of the current directory to the archive. A new archive"+
		"\n           is created if it does not already exist."+
		"\n           (Permitted switches: -c, -t:n)"+
		"\n     x    eXtract a snapshot from the archive to the current directory."+
		"\n           By default, the latest snapshot is selected; use the -s or -v switch"+
		"\n           to specify a different snapshot or volume."+
		"\n           (Permitted switches: -s:n, -v:n, -o)"+
		"\n     r    Restore a snapshot from the archive to the current directory."+
		"\n           Only files not identical to those already existing in the directory"+
		"\n           are extracted from the snapshot. The user is given the option of"+
		"\n           deleting existing files from the directory if they are not found in"+
		"\n           the snapshot. The latest snapshot is automatically selected; use"+
		"\n           the -s switch to specify a different snapshot."+
		"\n           (Permitted switches: -c, -t:n, -s:n, -o, -d)"+
		"\n     i    display Information on the archive"+
		"\n"+
		"\n  <Switches>:"+
		"\n    -c    include CRC-32 Checksums in file comparisons"+
		"\n    -t:n  use a +/- n-second allowance for last-modified Time in file"+
		"\n           comparisons (n is equal to or greater than zero)"+
		"\n    -s:n  extract files from Snapshot number n"+
		"\n    -v:n  extract files from Volume number n"+
		"\n    -o    automatically Overwrite existing files during extraction"+
		"\n    -d    automatically Delete existing files from the directory if they"+
		"\n           are not found in the snapshot"+
		"\n"+
		"\n  [Catalog]:"+
		"\n     File name of the catalog belonging to the ZipSnap archive"+
		"\n"+
		"\nNOTES:"+
		"\n"+
		"\n  1. A ZipSnap archive comprises a catalog (a text file) and a set of"+
		"\n      compressed volumes (Zip files). For each snapshot, only new or modified"+
		"\n      files are added to the ZipSnap archive; the catalog is updated (the"+
		"\n      previous file is renamed [catalog].n.bak), and an additional Zip file"+
		"\n      (named [catalog].yyyyMMdd-HHmmss.n.zip) is created if necessary."+
		"\n  2. When adding snapshots, ZipSnap creates compressed volumes that are in the"+
		"\n      same directory as the catalog. Similarly, when extracting snapshots or"+
		"\n      volumes, ZipSnap assumes that the compressed volumes can be found in the"+
		"\n      same directory as the catalog."+
		"\n  3. Files in a ZipSnap archive are identified by their relative path name,"+
		"\n      size, and last-modified time in seconds (by default, a +/- 1-second"+
		"\n      allowance is used for matching; use the -t switch to specify a different"+
		"\n      range). Additionally, the CRC-32 checksum can be used when comparing"+
		"\n      files (using the -c switch), but this consumes a significant amount of"+
		"\n      time and is unnecessary under most circumstances."+
		"\n"+
		"\nEXAMPLES:"+
		"\n"+
		"\n  1. Add a snapshot to an existing archive, or create a new archive if it"+
		"\n      does not already exist:" +
		"\n     java -jar ZipSnap.jar a \"C:\\My Work\\Backup\\ProjectAlpha\""+
		"\n"+
		"\n  2. Extract the latest snapshot from the archive, and overwrite existing" +
		"\n      files automatically:"+
		"\n     java -jar ZipSnap.jar x -o \"C:\\My Work\\Backup\\ProjectAlpha\""+
		"\n"+
		"\n  3. Extract snapshot number 2 from the archive:"+
		"\n     java -jar ZipSnap.jar x -s:2 \"C:\\My Work\\Backup\\ProjectAlpha\""+
		"\n"+
		"\n  4. Restore the latest snapshot from the archive, and overwrite existing"+
		"\n      files automatically:" +
		"\n     java -jar ZipSnap.jar r -o \"C:\\My Work\\Backup\\ProjectAlpha\""+
		"\n"+
		"\n  5. Display archive information:"+
		"\n     java -jar ZipSnap.jar i \"C:\\My Work\\Backup\\ProjectAlpha\""+
		"\n");
	}


	/**
	 * Main entry point.
	 */
	public static void main(String args[]) throws Exception {
		int i, iMax;

		p("\n" + progTitle + "\n");

		// process command-line arguments
		boolean invalidArg = false;
		String err = "";

		CheckArguments:
		do {//single iteration

			if (args.length<2) {
				invalidArg = true;
				err = "Insufficient arguments.";
				break CheckArguments;
			}

			String comd = args[0].trim().toLowerCase();
			argCatalog = args[args.length-1];

			if (comd.equals("a")) {
				iMax = args.length-1;
				for (i = 1; i < iMax; i++) {
					String sw = args[i].trim().toLowerCase();
					if (sw.equals("-c")) {
						checksumMode = Checksum.Use;
					} else if (sw.startsWith("-t:")) {
						argTimeAllowance = sw.substring(3);
					} else {
						invalidArg = true;
						err = "Invalid switch " + sw + " for command " + comd + ".";
						break CheckArguments;
					}
				}
				add();

			} else if (comd.equals("x")){
				iMax = args.length-1;
				for (i = 1; i < iMax; i++) {
					String sw = args[i].trim().toLowerCase();
					if (sw.startsWith("-s:")) {
						argSnapshotNumber = sw.substring(3);
					} else if (sw.startsWith("-v:")) {
						argVolumeNumber = sw.substring(3);
					} else if (sw.equals("-o")) {
						overwriteMode = Overwrite.Always;
					} else {
						invalidArg = true;
						err = "Invalid switch " + sw + " for command " + comd + ".";
						break CheckArguments;
					}
				}
				if ((argSnapshotNumber.length()>0) && (argVolumeNumber.length()>0)) {
					invalidArg = true;
					err = "Cannot use switches -s:" + argSnapshotNumber + " and -v:" + argVolumeNumber + " together for command " + comd + ".";
					break CheckArguments;
				}
				extract();

			} else if (comd.equals("r")){
				iMax = args.length-1;
				for (i = 1; i < iMax; i++) {
					String sw = args[i].trim().toLowerCase();
					if (sw.equals("-c")) {
						checksumMode = Checksum.Use;
					} else if (sw.startsWith("-t:")) {
						argTimeAllowance = sw.substring(3);
					} else if (sw.startsWith("-s:")) {
						argSnapshotNumber = sw.substring(3);
					} else if (sw.equals("-o")) {
						overwriteMode = Overwrite.Always;
					} else if (sw.equals("-d")) {
						deleteExistingMode = DeleteExisting.Always;
					} else {
						invalidArg = true;
						err = "Invalid switch " + sw + " for command " + comd + ".";
						break CheckArguments;
					}
				}
				restore();

			} else if (comd.equals("i")){
				iMax = args.length-1;
				for (i = 1; i < iMax; i++) {
					String sw = args[i].trim().toLowerCase();
					invalidArg = true;
					err = "Invalid switch " + sw + " for command " + comd + ".";
					break CheckArguments;
				}
				info();

			} else {
				invalidArg = true;
				err = "Invalid command: " + comd + ".";
				break CheckArguments;
			}
		} while (false);

		// invalid command-line arguments
		if (invalidArg) {
			p("\nERROR: " + err + "\n");
			printUsage();
			System.exit(10); // bad command-line arguments
		}

		System.exit(42); // unsuccessful termination (should never be reached)

	}


	/**
	 * Adds a snapshot to the ZipSnap archive.
	 */
	private static void add() {
		// declaration of commonly used variables (for speed)
		int i, iMax, j, jMax;
		File f = null;
		FileEntry fe = null;
		FileEntryComparator feComparator = new FileEntryComparator(checksumMode);

		// get current date and time
		String thisTime = (new SimpleDateFormat("yyyyMMdd-HHmmss")).format(new Date());

		try {
			directoryName = (new File("")).getCanonicalPath();
			directoryFile = new File(directoryName);
			catalogName = (new File(argCatalog)).getCanonicalPath();
			catalogFile = new File(catalogName);
			workName = (new File(catalogName)).getParentFile().getCanonicalPath(); // directory of the catalog
			workFile = new File(workName);

			p("\nCatalog  : [" + catalogName + "]");
			p("\nDirectory: [" + directoryName + "]");

			// check if Catalog is within the Directory
			if (catalogName.startsWith(directoryName+File.separator)) {
				p("\nERROR: The catalog cannot be contained within the directory.\nProgram aborted.\n");
				System.exit(1); // bad catalog location
			}
			// check if Directory is valid
			if (!(directoryFile.exists() && directoryFile.isDirectory() && directoryFile.canRead())) {
				p("\nERROR: The directory does not exist, is not a directory, or cannot be read.\nProgram aborted.\n");
				System.exit(2); // bad directory
			}
			// check if work (catalog) directory is valid
			if (!(workFile.exists() && workFile.isDirectory() && workFile.canRead() && workFile.canWrite())) {
				p("\nERROR: The catalog directory \"" + workName + "\"\ndoes not exist, is not a directory, cannot be read, or cannot be written to.\nProgram aborted.\n");
				System.exit(3); // bad catalog directory
			}

			// process time allowance
			int timeDelta[] = processTimeAllowance();

			// check if catalog exists
			boolean isNewArchive = false;
			if (catalogFile.exists()) {
				if (catalogFile.isFile() && catalogFile.canRead()) {
					// catalog exists and can be read, proceed to read
					isNewArchive = false;
					try {
						readCatalogFromFile();
					} catch (Exception e) {
						p("\nProgram aborted.\n");
						System.exit(4); // bad catalog
					}
					p("\n\nAdding snapshot to existing ZipSnap archive...");
				} else {
					// exists, but not file/readable
					p("\nERROR: The catalog exists, but is not a file, or cannot be read.\nProgram aborted.\n");
					System.exit(4); // bad catalog
				}
			} else {
				// catalog doesn't exist, so we create a new one
				isNewArchive = true;
				p("\n\nAdding snapshot to new ZipSnap archive...");
				snapshots = new ArrayList<SnapshotEntry>();
				volumes = new ArrayList<VolumeEntry>();
				files = new ArrayList<FileEntry>();
			}

			// create new snapshot, and add to archive
			SnapshotEntry se = new SnapshotEntry();
			se.time = thisTime;
			snapshots.add(se);
			int snapshotIndex = snapshots.size()-1;
			p("\nSnapshot: [" + (snapshotIndex+1) + ":" + thisTime + "]\n");

			// get directory contents
			if (checksumMode == Checksum.Use) p("\nNOTE: Computation of file checksums may take some time.");
			p("\nScanning directory...");
			ArrayList<FileEntry> directoryContents = null;
			try {
				directoryContents = getDirectoryContents(directoryFile);
			} catch (Exception e) {
				p("\nERROR: Problem scanning directory contents:\n" + e + "\nProgram aborted.\n");
				System.exit(2); // bad directory
			}
			p(" " + directoryContents.size() + " files & directories scanned.");

			// find out which files need to be added for this snapshot
			ArrayList<FileEntry> filesToAdd = new ArrayList<FileEntry>();

			// sort the files in the catalog first
			Collections.sort(files,feComparator);

			// step through each file in the directory, comparing it against the catalog
			iMax = directoryContents.size();

			FindNewFiles:
			for (i = 0; i < iMax; i++) {
				fe = directoryContents.get(i);

				// find the best match:
				// "among all files with matching NST(C), pick the one from the latest volume"
				int bestMatch = -1;
				long actualTime = fe.time;
				jMax = timeDelta.length;
				for (j = 0; j < jMax; j++) {
					fe.time = actualTime + timeDelta[j]; // in SECONDS
					int match = getBestMatch(fe, files, feComparator);
					if (match >= 0) bestMatch = match;
				}
				fe.time = actualTime;

				if (bestMatch >= 0) {
					fe = files.get(bestMatch);
					fe.snapshot.add(new Integer(snapshotIndex)); // add this snapshot number
				} else {
					filesToAdd.add(fe);
				}

			}
			p("\nNo. of new or modified files & directories to add: " + filesToAdd.size());

		    // create compressed zip volume, if necessary
		    if (filesToAdd.size()>0) {

				// create new volume
				VolumeEntry ve = new VolumeEntry();
				ve.time = thisTime;
				volumes.add(ve);
				int volumeIndex = volumes.size()-1;
				p("\nCompressed volume: [" + (volumeIndex+1) + ":" + thisTime + "]");

				String zipVolName = catalogName + "." + thisTime + "." + (volumeIndex+1) + ".zip";
				ArrayList<FileEntry> filesAdded = zipFiles(zipVolName, filesToAdd);

				long totalCompressedSize = (new File(zipVolName)).length();
				long totalSize = 0; 		// sum of uncompressed file sizes
				iMax = filesAdded.size();
				for (i = 0; i < iMax; i++) {
					fe = filesAdded.get(i);
					totalSize += fe.size;
			    	fe.volume = volumeIndex;
			    	fe.snapshot.add(new Integer(snapshotIndex));
					files.add(fe); // add file to archive catalog
				}

				p("\n  ------------");
				p("\n  No. of files & directories added: " + filesAdded.size() + " out of " + filesToAdd.size());
				p("\n  Total size           : " + totalSize + " bytes");
				p("\n  Total compressed size: " + totalCompressedSize + " bytes");
				p("\n  Compression ratio    : ");
				if (totalSize > 0) {
					p((int) (100.0 * (totalSize - totalCompressedSize) / totalSize) + "%");
				} else {
					p("-");
				}
			}

			// update catalog
			// as a precaution, rename the previous catalog first, if any
			if (isNewArchive) {
				p("\nWriting catalog...");
			} else {
				p("\nUpdating catalog...");
				// rename the previous catalog as backup
				if (!catalogFile.renameTo(new File(catalogName + "." + snapshotIndex + ".bak"))) {
					// cannot rename old catalog, abort program
					p("\nERROR: Unable to rename old catalog.\nThis snapshot has not been committed to the ZipSnap archive.\nProgram aborted.\n");
					System.exit(4); // bad catalog
				}
			}

			try {
				writeCatalogToFile();
			} catch (Exception e) {
				p("\nERROR: Problem writing to catalog file:\n" + e + "\nProgram aborted.\n");
				System.exit(4); // bad catalog
			}

			// successful termination
			p("\n\nDone!\n");
			System.exit(0);

		} catch (Exception e) {
			p("\nERROR: ZipSnap has encountered an error:\n" + e + "\nStack Trace:\n");
			e.printStackTrace();
			p("\nProgram aborted.\n");
			System.exit(42);

		}

	}


	/**
	 * Extracts a snapshot or volume from the ZipSnap archive.
	 */
	private static void extract() {
		// declaration of commonly used variables (for speed)
		int i, iMax;
		File f = null;
		FileEntry fe = null,fet = null;
		FileEntryComparator feComparator = new FileEntryComparator(checksumMode);

		try {
			directoryName = (new File("")).getCanonicalPath();
			directoryFile = new File(directoryName);
			catalogName = (new File(argCatalog)).getCanonicalPath();
			catalogFile = new File(catalogName);
			workName = (new File(catalogName)).getParentFile().getCanonicalPath(); // directory of the catalog
			workFile = new File(workName);

			p("\nCatalog  : [" + catalogName + "]");
			p("\nDirectory: [" + directoryName + "]");

			// check if directory is valid
			if (!(directoryFile.exists() && directoryFile.isDirectory() && directoryFile.canRead())) {
				p("\nERROR: The directory does not exist, is not a directory, or cannot be read.\nProgram aborted.\n");
				System.exit(2); // bad directory
			}
			// check if work (catalog) directory is valid
			if (!(workFile.exists() && workFile.isDirectory() && workFile.canRead())) {
				p("\nERROR: The catalog directory \"" + workName + "\"\ndoes not exist, is not a directory, or cannot be read.\nProgram aborted.\n");
				System.exit(3); // bad catalog directory
			}

			// check if catalog exists
			if (catalogFile.exists()) {
				if (catalogFile.isFile() && catalogFile.canRead()) {
					// catalog exists and can be read, proceed to read
					try {
						readCatalogFromFile();
					} catch (Exception e) {
						p("\nProgram aborted.\n");
						System.exit(4); // bad catalog
					}
				} else {
					// exists, but not file/readable
					p("\nERROR: The catalog exists, but is not a file, or cannot be read.\nProgram aborted.\n");
					System.exit(4); // bad catalog
				}
			} else {
				// catalog doesn't exist, so we abort
				p("\nERROR: The catalog does not exist.\nProgram aborted.\n");
				System.exit(4); // bad catalog
			}

			// process snapshot/volume number
			int maxSnapshotIndex = snapshots.size()-1;
			int snapshotIndex = -1;
			int maxVolumeIndex = volumes.size()-1;
			int volumeIndex = -1;

			ArrayList<ArrayList<FileEntry>> filesToExtract = new ArrayList<ArrayList<FileEntry>>();
			iMax = volumes.size();
			for (i = 0; i < iMax; i++) filesToExtract.add(new ArrayList<FileEntry>());
			int filesToExtractCount = 0;

			if (argVolumeNumber.length() == 0) {

				snapshotIndex = processSnapshotNumber(maxSnapshotIndex);
				p("\n\nExtracting " + ((snapshotIndex == maxSnapshotIndex) ? "latest " : "") + "snapshot from ZipSnap archive...");
				p("\nSnapshot: [" + (snapshotIndex+1) + ":" + snapshots.get(snapshotIndex).time + "]");

				iMax = files.size();
				for (i = 0; i < iMax; i++) {
					fe = files.get(i);
					if (fe.snapshot.contains(new Integer(snapshotIndex))) {
						filesToExtract.get(fe.volume).add(fe);
						filesToExtractCount++;
					}
				}

			} else {
				// extract specified volume number
				volumeIndex = processVolumeNumber(maxVolumeIndex);
				p("\n\nExtracting volume from ZipSnap archive...");
				p("\nCompressed volume: [" + (volumeIndex+1) + ":" + snapshots.get(volumeIndex).time + "]");
				iMax = files.size();
				for (i = 0; i < iMax; i++) {
					fe = files.get(i);
					if (fe.volume == volumeIndex) {
						filesToExtract.get(fe.volume).add(fe);
						filesToExtractCount++;
					}
				}

			}

			p("\nNo. of files & directories to extract: " + filesToExtractCount);

			iMax = volumes.size();
			ArrayList<FileEntry> filesExtracted = new ArrayList<FileEntry>();

			for (i = 0; i < iMax; i++) {
				ArrayList<FileEntry> filesToExtractFromVol = filesToExtract.get(i);
				if (filesToExtractFromVol.size() > 0) {

					String zipVolName = catalogName + "." + volumes.get(i).time + "." + (i+1) + ".zip";
					ArrayList<FileEntry> filesExtractedFromVol = unzipFiles(zipVolName, filesToExtractFromVol);
					filesExtracted.addAll(filesExtractedFromVol);

					p("\n  ------------");
					p("\n  No. of files & directories extracted: " +  filesExtractedFromVol.size() + " out of " +  filesToExtractFromVol.size());
				}
			}

			if (filesToExtractCount > 0) p("\nTotal no. of files & directories extracted: " +  filesExtracted.size() + " out of " + filesToExtractCount);

			// restore last-modified time
			if (filesExtracted.size() > 0) {
				p("\n\nRestoring last-modified time of extracted files & directories...");
				ArrayList<FileEntry> filesRestoredTime = restoreTime(filesExtracted);
				 p("\nTotal no. of file & directory times restored: " + filesRestoredTime.size() + " out of " + filesExtracted.size());
			 }

			// successful termination
			p("\n\nDone!\n");
			System.exit(0);

		} catch (Exception e) {
			p("\nERROR: ZipSnap has encountered an error:\n" + e + "\nStack Trace:\n");
			e.printStackTrace();
			p("\nProgram aborted.\n");
			System.exit(42);

		}

	 }


	/**
 	 * Restores a snapshot from the ZipSnap archive.
	 */
	private static void restore() {
		// declaration of commonly used variables (for speed)
		int i, iMax, j, jMax;
		File f = null;
		FileEntry fe = null,fet = null;
		FileEntryComparator feComparator = new FileEntryComparator(checksumMode);

		try {
			directoryName = (new File("")).getCanonicalPath();
			directoryFile = new File(directoryName);
			catalogName = (new File(argCatalog)).getCanonicalPath();
			catalogFile = new File(catalogName);
			workName = (new File(catalogName)).getParentFile().getCanonicalPath(); // directory of the catalog
			workFile = new File(workName);

			p("\nCatalog  : [" + catalogName + "]");
			p("\nDirectory: [" + directoryName + "]");

			// check if directory is valid
			if (!(directoryFile.exists() && directoryFile.isDirectory() && directoryFile.canRead())) {
				p("\nERROR: The directory does not exist, is not a directory, or cannot be read.\nProgram aborted.\n");
				System.exit(2); // bad directory
			}
			// check if work (catalog) directory is valid
			if (!(workFile.exists() && workFile.isDirectory() && workFile.canRead())) {
				p("\nERROR: The catalog directory \"" + workName + "\"\ndoes not exist, is not a directory, or cannot be read.\nProgram aborted.\n");
				System.exit(3); // bad catalog directory
			}

			// check if catalog exists
			if (catalogFile.exists()) {
				if (catalogFile.isFile() && catalogFile.canRead()) {
					// catalog exists and can be read, proceed to read
					try {
						readCatalogFromFile();
					} catch (Exception e) {
						p("\nProgram aborted.\n");
						System.exit(4); // bad catalog
					}
				} else {
					// exists, but not file/readable
					p("\nERROR: The catalog exists, but is not a file, or cannot be read.\nProgram aborted.\n");
					System.exit(4); // bad catalog
				}
			} else {
				// catalog doesn't exist, so we abort
				p("\nERROR: The catalog does not exist.\nProgram aborted.\n");
				System.exit(4); // bad catalog
			}

			// process time allowance
			int timeDelta[] = processTimeAllowance();

			// process snapshot number
			int maxSnapshotIndex = snapshots.size()-1;
			int snapshotIndex = processSnapshotNumber(maxSnapshotIndex);

			p("\n\nRestoring " + ((snapshotIndex == maxSnapshotIndex) ? "latest " : "") + "snapshot from ZipSnap archive...");
			p("\nSnapshot: [" + (snapshotIndex + 1) + ":" + snapshots.get(snapshotIndex).time + "]");

			// determine files in snapshot
			ArrayList<FileEntry> filesInSnapshot = new ArrayList<FileEntry>();
			int filesInSnapshotCount = 0;
			iMax = files.size();
			for (i = 0; i < iMax; i++) {
				fe = files.get(i);
				if (fe.snapshot.contains(new Integer(snapshotIndex))) {
					filesInSnapshot.add(fe);
					filesInSnapshotCount++;
				}
			}
			p("\nNo. of files & directories in snapshot: " + filesInSnapshotCount + "\n");

			// get directory contents
			if (checksumMode == Checksum.Use) p("\nNOTE: Computation of file checksums may take some time.");
			p("\nScanning directory...");
			ArrayList<FileEntry> directoryContents = null;
			try {
				directoryContents = getDirectoryContents(directoryFile);
			} catch (Exception e) {
				p("\nERROR: Problem scanning directory contents:\n" + e + "\nProgram aborted.\n");
				System.exit(2); // bad directory
			}
			p(" " + directoryContents.size() + " files & directories scanned.");

			// find out while files need to be extracted
			int filesToExtractCount = 0;
			ArrayList<ArrayList<FileEntry>> filesToExtract = new ArrayList<ArrayList<FileEntry>>();
			iMax = volumes.size();
			for (i = 0; i < iMax; i++) filesToExtract.add(new ArrayList<FileEntry>());

			// sort the files in directoryContents first
			Collections.sort(directoryContents, feComparator);

			// step through each file in the snapshot, comparing it against the directory contents
			FileEntryComparator feNameOnlyComparator = new FileEntryComparator();
			iMax = filesInSnapshot.size();
			StepFilesToExtract:
			for (i = 0; i < iMax; i++) {
				fe = filesInSnapshot.get(i);

				// find the best match:
				// "among all files with matching NST(C), pick the one from the latest volume"
				int bestMatch = -1;
				long actualTime = fe.time;
				jMax = timeDelta.length;
				for (j = 0; j < jMax; j++) {
					fe.time = actualTime + timeDelta[j]; // in SECONDS
					int match = getBestMatch(fe, directoryContents, feComparator);
					if (match >= 0) bestMatch = match;
				}
				fe.time = actualTime;

				if (bestMatch == -1) {
					// no match found, so we need to extract this file
					filesToExtract.get(fe.volume).add(fe);
					filesToExtractCount++;
				}

				// delete corresponding file from directoryContents, if any
				int match = getBestMatch(fe, directoryContents, feNameOnlyComparator);
				if (match >= 0) directoryContents.remove(match);

			}
			p("\nNo. of files & directories to extract: " + filesToExtractCount);
			// now, directoryContents contains files to be deleted


			iMax = volumes.size();
			ArrayList<FileEntry> filesExtracted = new ArrayList<FileEntry>();

			for (i = 0; i < iMax; i++) {
				ArrayList<FileEntry> filesToExtractFromVol = filesToExtract.get(i);
				if (filesToExtractFromVol.size() > 0) {

					String zipVolName = catalogName + "." + volumes.get(i).time + "." + (i+1) + ".zip";
					ArrayList<FileEntry> filesExtractedFromVol = unzipFiles(zipVolName, filesToExtractFromVol);
					filesExtracted.addAll(filesExtractedFromVol);

					p("\n  ------------");
					p("\n  No. of files & directories extracted: " +  filesExtractedFromVol.size() + " out of " +  filesToExtractFromVol.size());
				}
			}
			if (filesToExtractCount > 0) p("\nTotal no. of files & directories extracted: " +  filesExtracted.size() + " out of " + filesToExtractCount);

			// now, delete unaccounted files that exist in the directory but are not part of the snapshot
			if (directoryContents.size() > 0) {
				p("\n\nNo. of existing files & directories that are not part of the snapshot: " + directoryContents.size());
				ArrayList<FileEntry> filesDeleted = deleteFiles(directoryContents);
				 p("\nTotal no. of files & directories deleted: " + filesDeleted.size() + " out of " + directoryContents.size());
			 }

			// restore last-modified time
			if (filesExtracted.size() > 0) {
				p("\n\nRestoring last-modified time of extracted files & directories...");
				ArrayList<FileEntry> filesRestoredTime = restoreTime(filesExtracted);
				p("\nTotal no. of file & directory times restored: " + filesRestoredTime.size() + " out of " + filesExtracted.size());
			}

			// successful termination
			p("\n\nDone!\n");
			System.exit(0);

		} catch (Exception e) {
			p("\nERROR: ZipSnap has encountered an error:\n" + e + "\nStack Trace:\n");
			e.printStackTrace();
			p("\nProgram aborted.\n");
			System.exit(42);

		}

	}


	/**
	 * Displays information about the ZipSnap archive.
	 */
	private static void info() {
		int i, iMax, j, jMax;

		try {
			catalogName = (new File(argCatalog)).getCanonicalPath();
			catalogFile = new File(catalogName);

			p("\nCatalog: [" + catalogName + "]");

			// check if catalog exists
			if (catalogFile.exists()) {
				if (catalogFile.isFile() && catalogFile.canRead()) {
					// catalog exists and can be read
					try {
						readCatalogFromFile();
					} catch (Exception e) {
						p("\nProgram aborted.\n");
						System.exit(4); // bad catalog
					}
				} else {
					// exists, but not file/readable
					p("\nERROR: The catalog exists, but is not a file, or cannot be read.\nProgram aborted.\n");
					System.exit(4); // bad catalog
				}
			} else {
				// catalog doesn't exist, so we abort
				p("\nERROR: The catalog does not exist.\nProgram aborted.\n");
				System.exit(4); // bad catalog
			}

			// initialize counters
			int snapshotFilesCount[] = new int[snapshots.size()];
			iMax = snapshots.size();
			for (i = 0; i < iMax; i++) snapshotFilesCount[i] = 0;

			int snapshotFilesAddedCount[] = new int[snapshots.size()];
			iMax = snapshots.size();
			for (i = 0; i < iMax; i++) snapshotFilesAddedCount[i] = 0;

			int volumeFilesCount[] = new int[volumes.size()];
			iMax = volumes.size();
			for (i = 0; i < iMax; i++) volumeFilesCount[i] = 0;

			FileEntry fe = null;
			iMax = files.size();
			for (i = 0; i < iMax; i++) {
				fe = files.get(i);
				jMax = fe.snapshot.size();
				for (j = 0; j < jMax; j++) snapshotFilesCount[fe.snapshot.get(j).intValue()]++;
				snapshotFilesAddedCount[fe.snapshot.get(0).intValue()]++;
				volumeFilesCount[fe.volume]++;
			}

			String pad = null;
			p("\n\nThis ZipSnap archive contains " + snapshots.size() + " snapshots compressed over " + volumes.size() + " volumes.");
			p("\n\nSnapshots:");
			p("\n                                 No. of files & directories");
			p("\n    Number         Time                 (total/added)");
			p("\n   --------  -----------------  ----------------------------");

			iMax = snapshots.size();
			for (i = 0; i < iMax; i++) {
				pad = (i + 1 < 10) ? " " : "";
				pad += (i + 1 < 100) ? " " : "";
				p("\n     " + pad + (i + 1) + "      " + snapshots.get(i).time + "           " + snapshotFilesCount[i] + "/" + snapshotFilesAddedCount[i]);
			}

			p("\n\nVolumes:\n");
			p("\n    Number         Time          No. of files & directories");
			p("\n   --------  -----------------  ----------------------------");

			iMax = volumes.size();
			for (i = 0; i < iMax; i++) {
				pad = (i + 1 < 10) ? " " : "";
				pad += (i + 1 < 100) ? " " : "";
				p("\n     " + pad + (i + 1) + "      " + volumes.get(i).time + "             " + volumeFilesCount[i]);
			}

			// successful termination
			p("\n\n");
			System.exit(0);

		} catch (Exception e) {
			p("\nERROR: ZipSnap has encountered an error:\n" + e + "\nStack Trace:");
			e.printStackTrace();
			p("\nProgram aborted.\n");
			System.exit(42);

		}

	}


	/**
	 * Writes the catalog (files, snapshots, volumes) to a file.
	 */
	private static void writeCatalogToFile() throws Exception {
		int i, iMax, j, jMax;
		SnapshotEntry se = null;
		VolumeEntry ve = null;
		FileEntry fe = null;

		PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(catalogFile)));

		pw.println("# " + progTitle);
		pw.println("# IMPORTANT: PLEASE DO NOT MODIFY THIS CATALOG MANUALLY!");

		pw.println("\n# Number of Snapshots");
		pw.println("ns:" + snapshots.size());
		pw.println("# SNAPSHOT:index:time");
		iMax = snapshots.size();
		for (i = 0; i < iMax; i++) {
			se = snapshots.get(i);
			pw.println("s:" + (i+1) + ":" + se.time);
		}

		pw.println("\n# Number of Volumes");
		pw.println("nv:" + volumes.size());
		pw.println("# VOLUME:index:time");
		iMax = volumes.size();
		for (i = 0; i < iMax; i++) {
			ve = volumes.get(i);
			pw.println("v:" + (i+1) + ":" + ve.time);
		}

		Collections.sort(files,new FileEntryComparator(checksumMode)); // sort the files first
		pw.println("\n# Number of Files & Directories");
		pw.println("nf:" + files.size());
		pw.println("# FILE:index:[name],size,time(in sec),checksum,volume,snapshot");
		iMax = files.size();
		for (i = 0; i < iMax; i++) {
			fe = files.get(i);
			pw.print("f:" + (i+1) + ":[" + nativeToNeutral(fe.name) + "]," + fe.size + "," + fe.time + "," + fe.checksum + "," + (fe.volume+1));
			jMax = fe.snapshot.size();
			for (j = 0; j < jMax; j++) {
				pw.print("," + (fe.snapshot.get(j).intValue()+1));
			}
			pw.println("");
		}

		pw.println("\n# End-of-Catalog\n");
		pw.flush();
		pw.close();
	}


	/**
	 * Reads the ZipSnap catalog from a file.
  	 */
	private static void readCatalogFromFile() throws Exception {
		int i, iMax;
		int lineno = 0;
		String line = null;

		try {
			// initialization
			int ns = -1,nv = -1,nf = -1;
			snapshots = new ArrayList<SnapshotEntry>();
			volumes = new ArrayList<VolumeEntry>();
			files = new ArrayList<FileEntry>();

			String s = null;
			int index;
			SnapshotEntry se = null;
			VolumeEntry ve = null;
			FileEntry fe = null;

			BufferedReader br = new BufferedReader(new FileReader(catalogFile));

			while (br.ready()) {
				lineno++;
				line = br.readLine().trim();
				s = line;

				if (s.startsWith("ns:")) {
					// number of snapshots
					ns = Integer.parseInt(s.substring(3).trim());
					for (i = 0; i < ns; i++) snapshots.add(null);

				} else if (s.startsWith("nv:")) {
					// number of volumes
					nv = Integer.parseInt(s.substring(3).trim());
					for (i = 0; i < nv; i++) volumes.add(null);

				} else if (s.startsWith("nf:")) {
					// number of files
					nf = Integer.parseInt(s.substring(3).trim());
					for (i = 0; i < nf; i++) files.add(null);

				} else if (s.startsWith("s:")) {
					// SNAPSHOT:index:time
					se = new SnapshotEntry();
					s = s.substring(2).trim();
					index = Integer.parseInt(s.substring(0,s.indexOf(":")).trim())-1;
					se.time = s.substring(s.indexOf(":")+1).trim();
					snapshots.set(index,se);

				} else if (s.startsWith("v:")) {
					// VOLUME:index:time
					ve = new VolumeEntry();
					s = s.substring(2).trim();
					index = Integer.parseInt(s.substring(0,s.indexOf(":")).trim())-1;
					ve.time = s.substring(s.indexOf(":")+1).trim();
					volumes.set(index,ve);

				} else if (s.startsWith("f:")) {
					// FILE:index:[name],size,time(in sec),checksum,volume,snapshot
					fe = new FileEntry();
					s = s.substring(2).trim();
					index = Integer.parseInt(s.substring(0,s.indexOf(":")).trim())-1;
					fe.name = neutralToNative(s.substring(s.indexOf("[")+1,s.lastIndexOf("]")));
					s = s.substring(s.lastIndexOf("]")+1).trim();
					s = s.substring(s.indexOf(",")+1).trim();
					String t[] = s.split(",");
					fe.size = Long.parseLong(t[0].trim());
					fe.time = Long.parseLong(t[1].trim());
					fe.checksum = Long.parseLong(t[2].trim());
					fe.volume = Integer.parseInt(t[3].trim())-1;
					iMax = t.length;
					for (i = 4; i < iMax; i++) fe.snapshot.add(new Integer(Integer.parseInt(t[i].trim())-1));
					files.set(index,fe);

				}

			}

			br.close();
			// check the counts
			if ((snapshots.size() != ns) ||
				(volumes.size() != nv) ||
			  (files.size() != nf)) {
				p("\nERROR: Incorrect number of entries in catalog.");
				throw new Exception();
			}

		} catch (Exception e) {
			p("\nERROR: Unable to parse catalog:\n" + e);
			p("\nLine " + lineno + ": " + line);
			throw e;
		}
	}


	/**
	 * Adds a given list of files to the specified Zip file.
	 *
	 * @params zipFilename The file name of the Zip file.
	 * @params filesToZip The ArrayList of FileEntry's to be added to the Zip file.
	 * @returns The ArrayList of FileEntry's successfully added to the Zip file.
	 */
	private static ArrayList<FileEntry> zipFiles(String zipFilename, ArrayList<FileEntry> filesToZip) {

		ArrayList<FileEntry> filesZipped = new ArrayList<FileEntry>(); // return value
		if (filesToZip.size() == 0) return filesZipped;

	    p("\n\nCreating Zip file " + zipFilename + "...");

	    // set up the compressed output stream to write to a Zip file
	    ZipOutputStream zos = null;
	    try {
		    zos = new ZipOutputStream(new FileOutputStream(zipFilename));
			zos.setLevel(9); // set to max compression
		} catch (Exception e) {
			p("\nERROR: Unable to create Zip file.\n" + e + "\nProgram aborted.\n");
			System.exit(7);
		}

		boolean isDeflated;
		FileEntry fe = null;
		ZipEntry ze = null;
		FileInputStream fis = null;
		int i;
		int iMax = filesToZip.size();
		int byteCount;

		StepAddFilesToZip:
		for (i = 0; i < iMax; i++) {

			try {
		    	fe = filesToZip.get(i);
				p("\n  Adding: " + fe.name);

		    	// populate ZipEntry properties
		    	ze = new ZipEntry(nativeToNeutral(fe.name));
		    	ze.setSize(fe.size);
		  		ze.setTime(fe.time * 1000); // in MILLISECONDS

		    	// decide on compression method
		    	isDeflated = (fe.size < DEFLATED_MIN) ? false : true;

		    	// continue populating ZipEntry properties
		    	if (isDeflated) {
			    	ze.setMethod(ZipEntry.DEFLATED);

				} else {
		  			ze.setMethod(ZipEntry.STORED);

		    		// need to compute CRC32 checksum before STORING ZipEntry
		    		if (checksumMode == Checksum.Ignore) {
			    		try {
				    		fe.checksum = computeFileChecksum(new File(fe.name));
			    		} catch (Exception e) {
							p("\n  WARNING: Unable to compute CRC-32 checksum. This " + (fe.isFile() ? "file" : "directory") + " will not be cataloged.");
							continue StepAddFilesToZip; // proceed to next File i
			    		}
		    		}

		    		ze.setCrc(fe.checksum);
		  		}

		  		// commit Zip Entry to the zip file
		  		zos.putNextEntry(ze);

		  		// read the file and compress the data; ignore if it is a directory
	  			if (fe.isFile()) {
		    		fis = new FileInputStream(fe.name);
		    		while ((byteCount = fis.read(byteBuffer, 0, BUFFER_SIZE)) != -1) {
		    			zos.write(byteBuffer, 0, byteCount);
	    		 	}
			    	fis.close();
			    }

			    // close the zip stream and prepare for next entry
		    	zos.closeEntry();

		    	if (isDeflated) {
			    	fe.checksum = ze.getCrc();
			    	p(" (" + (int) (100.0 * (ze.getSize() - ze.getCompressedSize()) / ze.getSize()) + "%)");
				}

			} catch (Exception e) {
	  			p("\n  WARNING: Unable to add to Zip file. This " + (fe.isFile() ? "file" : "directory") + " will not be cataloged.");
				continue StepAddFilesToZip; // proceed to next File i
			}

	    	// take note of successfully written FileEntry
	    	filesZipped.add(fe);
		}

		// close the zip stream
		try {
			zos.flush();
			zos.close();
		} catch (Exception e) {
			p("\nERROR: Unable to close Zip file.\n" + e + "\nProgram aborted.\n");
			System.exit(7);
		}

		// return ArrayList of successfully written FileEntry's
		return filesZipped;
	}


	/**
	 * Extracts a given list of files from the specified Zip file.
	 *
	 * @params zipFilename The file name of the Zip file.
	 * @params filesToUnzip The ArrayList of FileEntry's to be extracted from the Zip file.
	 * @returns The ArrayList of FileEntry's successfully extracted from the Zip file.
	 */
	private static ArrayList<FileEntry> unzipFiles(String zipFilename, ArrayList<FileEntry> filesToUnzip) {

		ArrayList<FileEntry> filesUnzipped = new ArrayList<FileEntry>(); // return value
		if (filesToUnzip.size() == 0) return filesUnzipped;

	    p("\n\nReading Zip file " + zipFilename + "...");

		File zipFile = new File(zipFilename);

		if (!(zipFile.exists() && zipFile.isFile() && zipFile.canRead())) {
			p("\n  WARNING: Zip file cannot be found or read. Files & directories from this volume will not be extracted.");
			return filesUnzipped;
		}

		// zip file exists and can be read
		ZipFile zf = null;
		File f = null;
		FileEntry fe = null;
		ZipEntry ze = null;
		InputStream zis = null;
		FileOutputStream fos = null;
		int i;
		int iMax = filesToUnzip.size();
		int byteCount;

		try {
			zf = new ZipFile(zipFile);
		} catch (Exception e) {
			p("\n  WARNING: Zip file cannot be opened. Files & directories from this volume will not be extracted.");
			return filesUnzipped;
		}

		StepExtractFilesFromZip:
		for (i = 0; i < iMax; i++) {

			fe = filesToUnzip.get(i);
			f = new File(fe.name);
			p("\n  Extracting: " + fe.name);

			try {

				if (fe.isFile()) {

					// extracting a file
					boolean extractFile = false;

					if (f.exists() && f.isFile()) {

						// overwrite file with file?
						p("\n    WARNING: File already exists.");
						if (overwriteMode == Overwrite.Always) {
							p(" File overwritten.");
							extractFile = true;
						} else if (overwriteMode == Overwrite.Never) {
							p(" File skipped.");
							extractFile = false;
						} else if (overwriteMode == Overwrite.Ask) {
							p("\n");
							UserPrompt.Reply validReplies[] = { UserPrompt.Reply.Yes, UserPrompt.Reply.No, UserPrompt.Reply.Always, UserPrompt.Reply.Never};
							UserPrompt.Reply reply = UserPrompt.prompt("    Overwrite file? ", validReplies);
							if (reply == UserPrompt.Reply.Yes) {
								extractFile = true;
							} else if (reply == UserPrompt.Reply.No) {
								extractFile = false;
							} else if (reply == UserPrompt.Reply.Always) {
								extractFile = true;
								overwriteMode = Overwrite.Always;
							} else if (reply == UserPrompt.Reply.Never) {
								extractFile = false;
								overwriteMode = Overwrite.Never;
							}
						}

					} else if (f.exists() && f.isDirectory()) {

						// problem: cannot overwrite directory with file
						p("\n    WARNING: A directory of the same name exists. File will not be extracted.");
						extractFile = false;

					} else {

						// file doesn't already exist, so we just unzip it
						extractFile = true;

					}

					if (extractFile) {

						ze = zf.getEntry(nativeToNeutral(fe.name));
						if (ze == null) {

							p("\n    WARNING: " + (fe.isFile() ? "File" : "Directory") + " not found in Zip file, and will not be extracted.");
							continue StepExtractFilesFromZip; // proceed to the next file i to extract

						} else {

							if (f.getParentFile() != null) {
								f.getParentFile().mkdirs();
								if (!(f.getParentFile().exists() && f.getParentFile().isDirectory())) {
									// unsuccessful
									p("\n    WARNING: Unable to create parent directories for file. File skipped.");
									continue StepExtractFilesFromZip; // proceed to next File i
								}
							}

							try {
								// write uncompressed data to file
								zis = zf.getInputStream(ze);
								fos = new FileOutputStream(f);
								while ((byteCount = zis.read(byteBuffer, 0, BUFFER_SIZE)) != -1) {
									fos.write(byteBuffer, 0, byteCount);
								}
						    	fos.flush();
								fos.close();
								zis.close();
							} catch (Exception e) {
								// unsuccessful
								p("\n    WARNING: Unable to extract file.");
								continue StepExtractFilesFromZip; // proceed to next File i
							}
							// successfully extracted the file
							filesUnzipped.add(fe);
						}
					}

				} else if (fe.isDirectory()) {

					// extracting a directory
					boolean extractDirectory = false;

					if (f.exists() && f.isDirectory()) {

						// overwrite directory with directory? (update time only)
						p("\n    WARNING: Directory already exists.");
						if (overwriteMode == Overwrite.Always) {
							p(" Directory overwritten.");
							extractDirectory = true;
						} else if (overwriteMode == Overwrite.Never) {
							p(" Directory skipped.");
							extractDirectory = false;
						} else if (overwriteMode == Overwrite.Ask) {
							p("\n");
							UserPrompt.Reply validReplies[] = { UserPrompt.Reply.Yes, UserPrompt.Reply.No, UserPrompt.Reply.Always, UserPrompt.Reply.Never};
							UserPrompt.Reply reply = UserPrompt.prompt("    Overwrite directory? ", validReplies);
							if (reply == UserPrompt.Reply.Yes) {
								extractDirectory = true;
							} else if (reply == UserPrompt.Reply.No) {
								extractDirectory = false;
							} else if (reply == UserPrompt.Reply.Always) {
								extractDirectory = true;
								overwriteMode = Overwrite.Always;
							} else if (reply == UserPrompt.Reply.Never) {
								extractDirectory = false;
								overwriteMode = Overwrite.Never;
							}
						}

					} else if (f.exists() && f.isFile()) {

						// problem: cannot overwrite file with directory
						p("\n    WARNING: A file of the same name exists. Directory will not be extracted.");
						extractDirectory = false;

					} else {

						// directory doesn't exist yet, we we just create it
						extractDirectory = true;

					}

					if (extractDirectory) {

						f.mkdirs();
						if (f.exists() && f.isDirectory()) {

							// successfully created the directory
							filesUnzipped.add(fe);

						} else {

							// unsuccessful
							p("\n    WARNING: Unable to extract directory.");
							continue StepExtractFilesFromZip; // proceed to next File i

						}
					}
				}
			} catch (Exception e) {

				p("\n    WARNING: Unable to extract " + (fe.isFile() ? "file" : "directory") + ".");
				continue StepExtractFilesFromZip; // proceed to next File i

			}

	    }

		// close the zip file
		try {
			zf.close();
		} catch (Exception e) {
			p("\n  WARNING: Unable to close Zip file.\n");
		}

		// return ArrayList of successfully unzipped FileEntry's
		return filesUnzipped;

	}


	/**
	 * Restores the last-modified time of the given list of files
	 *
	 * @params filesToRestoreTime The ArrayList of FileEntry's corresponding to the files.
	 * @returns The ArrayList of FileEntry's whose time is successfully restored.
	 */
	private static ArrayList<FileEntry> restoreTime(ArrayList<FileEntry> filesToRestoreTime) {
		ArrayList<FileEntry> filesRestoredTime = new ArrayList<FileEntry>(); // return value
		if (filesToRestoreTime.size() == 0) return filesRestoredTime;

		// sort the list of files by name only
		Collections.sort(filesToRestoreTime, new FileEntryComparator());

		File f = null;
		FileEntry fe = null;

		// step through each file in reverse-order
		for (int i = filesToRestoreTime.size() - 1; i >= 0; i--) {
			fe = filesToRestoreTime.get(i);
			f = new File(fe.name);
			if (f.exists() && (f.isFile()==fe.isFile()) && (f.isDirectory()==fe.isDirectory())) {
				if (f.setLastModified(fe.time*1000)) { // in MILLISECONDS
					// successfully restored
					filesRestoredTime.add(fe);
				} else {
					p("\n  WARNING: Unable to restore last-modified time stamp of " + fe.name);
				}
			}
		}

		return filesRestoredTime;
	}


	/**
	 * Deletes the given list of files
	 *
	 * @params filesToDelete The ArrayList of FileEntry's corresponding to the files.
	 * @returns The ArrayList of FileEntry's successfully deleted.
	 */
	private static ArrayList<FileEntry> deleteFiles(ArrayList<FileEntry> filesToDelete) {
		ArrayList<FileEntry> filesDeleted = new ArrayList<FileEntry>(); // return value
		if (filesToDelete.size() == 0) return filesDeleted;

		// sort the list of files by name only
		Collections.sort(filesToDelete,new FileEntryComparator());

		File f = null;
		FileEntry fe = null;
		boolean deleteFile;

		// step through each file in reversed order
		StepDeleteFile:
		for (int i = filesToDelete.size() - 1; i >= 0; i--) {
			fe = filesToDelete.get(i);
			f = new File(fe.name);
			deleteFile = false;

			if (deleteExistingMode == DeleteExisting.Always) {
				p("\n  Deleting: " + fe.name);
				deleteFile = true;
			} else if (deleteExistingMode == DeleteExisting.Never) {
				p("\n  Skipping: " + fe.name);
				deleteFile = false;
			} else if (deleteExistingMode == DeleteExisting.Ask) {
				p("\n");
				UserPrompt.Reply validReplies[] = { UserPrompt.Reply.Yes, UserPrompt.Reply.No, UserPrompt.Reply.Always, UserPrompt.Reply.Never};
				UserPrompt.Reply reply = UserPrompt.prompt("  Delete: " + fe.name + " ? ", validReplies);
				if (reply == UserPrompt.Reply.Yes) {
					deleteFile = true;
				} else if (reply == UserPrompt.Reply.No) {
					deleteFile = false;
				} else if (reply == UserPrompt.Reply.Always) {
					deleteFile = true;
					deleteExistingMode = DeleteExisting.Always;
				} else if (reply == UserPrompt.Reply.Never) {
					deleteFile = false;
					deleteExistingMode = DeleteExisting.Never;
				}
			}

			if (deleteFile && f.exists() && (f.isFile()==fe.isFile())
					&& (f.isDirectory()==fe.isDirectory())) {
				if (f.delete()) {
					// successfully deleted
					filesDeleted.add(fe);
				} else {
					p("\n    WARNING: Unable to delete " + (fe.isFile() ? "file" : "directory") + ".");
				}
			}
		}

		return filesDeleted;
	}


	/**
	 * Computes the CRC-32 checksum of a file
	 *
	 * @params f The File whose checksum is to be computed
	 * @returns The CRC-32 checksum value.
	 */
	private static long computeFileChecksum(File f) throws Exception {
		if (f.isFile()) {
			CRC32 crc32 = new CRC32();
			FileInputStream fis = new FileInputStream(f);
			int byteCount;
			while ((byteCount = fis.read(byteBuffer, 0, BUFFER_SIZE)) != -1) {
				crc32.update(byteBuffer, 0, byteCount);
			}
			fis.close();
			return crc32.getValue();
		} else {
			return 0;
		}
	}


	/**
	 * Crawls a directory to gather its files and subdirectories.
	 *
	 * @params directoryFile The File representing the directory to crawl.
	 * @returns ArrayList of FileEntry's representing the files and subdirectories
	 * 	of directoryFile.
	 */
	private static ArrayList<FileEntry> getDirectoryContents(File directoryFile) throws Exception {

		ArrayList<FileEntry> directoryContents = new ArrayList<FileEntry>(); // return value
		File listFiles[] = directoryFile.listFiles(); // files and directories
		File f = null;
		FileEntry fe = null;
		int i;
		int iMax = listFiles.length;

		// step through directory contents to add files & directories
		StepDirectoryContents:
		for (i = 0; i < iMax; i++) {
			f = listFiles[i];
			fe = new FileEntry();
			fe.name = f.getCanonicalPath().substring(directoryName.length() + 1); // get relative path name

			if (f.isDirectory() && !fe.name.endsWith(File.separator)) {
				fe.name += File.separator; // add trailing slash for directories
			}

			if (fe.name.contains("\n")) {
				p("\n  WARNING: Relative path name of file [" + fe.name + "] contains '\n' character.\nThis file will be ignored.");
				continue StepDirectoryContents; // proceed to next File i
			}

			fe.size = f.length();
			fe.time = f.lastModified() / 1000; // in SECONDS

			if (checksumMode == Checksum.Use) {
				try {
					fe.checksum = computeFileChecksum(f);
				} catch (Exception e) {
					p("\n  WARNING: Cannot compute CRC-32 checksum for file [" + fe.name + "].\nThis file will be ignored.");
					continue StepDirectoryContents; // proceed to next File i
				}
			}

			directoryContents.add(fe);
		}

		// now, recurse into subdirectories
		for (i = 0; i < iMax; i++) {
			f = listFiles[i];
			if (f.isDirectory()) {
				directoryContents.addAll(getDirectoryContents(f));
			}
		}

		return directoryContents;
	}


	/**
	 * Gets the closest match for a given file, from a given list, using a given comparison.
	 * feList MUST be sorted before executing this method.
	 *
	 * @params fe The FileEntry to search for.
	 * @params feList The ArrayList of FileEntry's to search in.
	 * @params feComparator The comparison to use for identifying matches.
	 * @returns The index of the best match in feList.
     */
	private static int getBestMatch(FileEntry fe, ArrayList<FileEntry> feList, FileEntryComparator feComparator) {
		// ensure the files list is sorted before doing binary search

		int match = Collections.binarySearch(feList, fe, feComparator);
		if (match < 0) return -1; // no match found, return

		// a match found, now look for the BEST match (the one with the latest volume)
		int bestMatch = match;
		int latestVolume = Integer.MIN_VALUE;
		int i;
		int iMax = feList.size();
		FileEntry fei = null;

		StepUpList:
		for (i = match; i >= 0; i--) {
			fei = feList.get(i);
			if (feComparator.compare(fe, fei) == 0) {
				// match found
				if (fei.volume > latestVolume) {
					latestVolume = fei.volume;
					bestMatch = i;
				}
			} else {
				// not matching
				break StepUpList;
			}
		}

		StepDownList:
		for (i = match; i < iMax; i++) {
			fei = feList.get(i);
			if (feComparator.compare(fe, fei) == 0) {
				// match found
				if (fei.volume > latestVolume) {
					latestVolume = fei.volume;
					bestMatch = i;
				}
			} else {
				// not matching
				break StepDownList;
			}
		}

		return bestMatch;
	}


	/**
	 * Creates an array of time allowance values.
	 *
	 * @returns The int array containing the time allowance values, arranged
	 *  from least to most desired.
	 */
	private static int[] processTimeAllowance() {

		int timeAllowance = 1; // defaults to +/- 1-second allowance
		boolean validTimeAllowance = true;

		if (argTimeAllowance.length() > 0) {
			try {
				timeAllowance = Integer.parseInt(argTimeAllowance);
				if (timeAllowance < 0) validTimeAllowance = false;
			} catch (Exception e) {
				validTimeAllowance = false;
			}
			if (!validTimeAllowance) {
				p("\nERROR: Invalid allowance of " + argTimeAllowance + " seconds for last-modified time.\nPlease specify an integer equal to or greater than 0.\nProgram aborted.\n");
				System.exit(10); // bad command-line arguments
			}
		}

		// get range of time deltas (last-modified time allowance)
		// arranged from least desired to most desired; values in SECONDS
		int timeDelta[] = new int[1 + 2 * timeAllowance];
		int j = 0;
		for (int i = timeAllowance; i > 0; i--) {
			timeDelta[j++] = -i;
			timeDelta[j++] = i;
		}
		timeDelta[j++] = 0; // most desired delta = 0 (no difference in time)

		return timeDelta;
	}


	/**
	 * Processes the command-line argument for the Snapshot Number.
	 *
	 * @returns The corresponding snapshot index in snapshots
	 */
	private static int processSnapshotNumber(int maxSnapshotIndex) {

		int snapshotIndex = maxSnapshotIndex; // default value
		boolean validSnapshotIndex = true;

		if (argSnapshotNumber.length() > 0) {

			try {
				snapshotIndex = Integer.parseInt(argSnapshotNumber) - 1;
				if ((snapshotIndex < 0) || (snapshotIndex > maxSnapshotIndex)) validSnapshotIndex = false;
			} catch (Exception e) {
				validSnapshotIndex = false;
			}

			if (!validSnapshotIndex) {
				p("\nERROR: Invalid snapshot number " + argSnapshotNumber + " for this ZipSnap archive.\nPlease specify a snapshot number from 1 to " +  (maxSnapshotIndex + 1) + ".\nProgram aborted.\n");
				System.exit(10); // bad command-line
			}
		}

		return snapshotIndex;
	}


	/**
	 * Processes the command-line argument for the Volume Number.
	 *
	 * @returns The corresponding volume index in volumes
	 */
	private static int processVolumeNumber(int maxVolumeIndex) {

		int volumeIndex = maxVolumeIndex; // default value
		boolean validVolumeIndex = true;

		if (argVolumeNumber.length() > 0) {

			try {
				volumeIndex = Integer.parseInt(argVolumeNumber) - 1;
				if ((volumeIndex < 0) || (volumeIndex > maxVolumeIndex)) validVolumeIndex = false;
			} catch (Exception e) {
				validVolumeIndex = false;
			}

			if (!validVolumeIndex) {
				p("\nERROR: Invalid volume number " + argVolumeNumber + " for this ZipSnap archive.\nPlease specify a volume number from 1 to " +  (maxVolumeIndex + 1) + ".\nProgram aborted.\n");
				System.exit(10); // bad command-line
			}
		}

		return volumeIndex;
	}


	/**
	 * Converts a file name from native form to neutral form.
	 *
	 * @params s The native file name to be converted.
	 * @returns The neutral file name.
     */
	private static String nativeToNeutral(String s) {
		return s.replace(File.separator,NAME_SEPARATOR);
	}


	/**
	 * Converts a file name from neutral form to native form.
	 *
	 * @params s The neutral file name to be converted.
	 * @returns The native file name.
     */
	private static String neutralToNative(String s) {
		return s.replace(NAME_SEPARATOR,File.separator);
	}


	/**
	 * General purpose method for printing to standard output.
	 *
	 * @params o The object to be printed.
     */
	private static void p(Object o)  {
		System.out.print(o + "");
	}


	/**
	 * Defines the way two FileEntry's are to be compared.
     */
	static class FileEntryComparator implements Comparator<FileEntry> {
		private boolean useChecksum;
		private boolean useNameOnly;

		/** constructor with the option of using checksum comparison */
		FileEntryComparator(Checksum mode) {
			useChecksum = (mode == Checksum.Use);
			useNameOnly = false;
		}

		/** constructor for plain name comparison only */
		FileEntryComparator() {
			useNameOnly = true;
		}

		/** comparison based on: name, size, time, and checksum (optional) */
		public int compare(FileEntry o1, FileEntry o2) {
			// returns "o1-o2"
			int i = 0;
			long l = 0;

			// compare name's
			i = o1.name.compareTo(o2.name);
			if (i != 0) return i;

			// end-of-comparison if only name is used
			if (useNameOnly) return 0;

			// compare size's
			l = o1.size - o2.size;
			if (l != 0) return (l < 0) ? -1 : 1;

			// compare time's
			l = o1.time - o2.time;
			if (l != 0) return (l < 0) ? -1 : 1;

			// end-of-comparison if checksum is not used
			if (!useChecksum) return 0;

			// compare time's
			l = o1.checksum - o2.checksum;
			if (l != 0) return (l < 0) ? -1 : 1;

			// end-of-comparison
			return 0;
		}
	}


	/**
	 * Defines a file with all the relevant attributes.
     */
	static class FileEntry {
		String name; 	// if a name has a trailing separator char, the FileEntry describes a directory
		long size; 		// size of file in BYTES
		long time; 		// last-modified time in SECONDS since the epoch (00:00:00 GMT, January 1, 1970)
		long checksum; 	// CRC-32 checksum
		int volume; 		// which Zip volume is this file/directory stored in?
		ArrayList<Integer> snapshot; // which snapshot(s) is this file/directory a member of?

		FileEntry() {
			this.name = null;
			this.size = -1;
			this.time = -1;
			this.checksum = -1;
			this.volume = -1;
			this.snapshot = new ArrayList<Integer>();
		}

		boolean isDirectory() {
			return name.endsWith(File.separator);
		}

		boolean isFile() {
			return (!name.endsWith(File.separator));
		}

	}


	/**
	 * Defines a snapshot.
     */
	static class SnapshotEntry {
		private String time; // yyyyMMdd-HHmmss
	}


	/**
	 * Defines a volume.
     */
	static class VolumeEntry {
		private String time; // yyyyMMdd-HHmmss
	}


	/**
	 * Defines class for obtaining keyboard input from the user.
	 */
	static class UserPrompt {
		public enum Reply {Yes, No, Always, Never};

		/**
		 * Repeatedly prompts a user with a specified message, and processes the
		 * user's keyboard response until a valid reply is received. Keyboard
		 * inputs are treated in a case-INsensitive manner.
		 *
		 * @params message The message string to print as a prompt to the user.
		 * @params validReplies An array of valid UserPromptReply's.
		 * @returns The valid UserPromptReply received from the user.
	     */
		 public static UserPrompt.Reply prompt(String message, UserPrompt.Reply validReplies[]) {
			try {
				BufferedReader kb = new BufferedReader(new InputStreamReader(System.in));
				String choice = "";
				int i, iMax = validReplies.length;
				while (true) {
					p(message);
					for (i = 0; i < iMax; i++) {
						if (validReplies[i] == UserPrompt.Reply.Yes) {
							p("(Y)es");
						} else if (validReplies[i] == UserPrompt.Reply.No) {
							p("(N)o");
						} else if (validReplies[i] == UserPrompt.Reply.Always) {
							p("(A)lways");
						} else if (validReplies[i] == UserPrompt.Reply.Never) {
							p("neve(R)");
						}
						if (i < iMax - 1) p("/");
					}
					p(": ");
					choice = kb.readLine().trim();
					for (i = 0; i < iMax; i++) {
						if ((validReplies[i] == UserPrompt.Reply.Yes) &&
								(choice.equalsIgnoreCase("y") || choice.equalsIgnoreCase("yes") || choice.equalsIgnoreCase("(y)es"))) {
							return UserPrompt.Reply.Yes;

						} else if ((validReplies[i] == UserPrompt.Reply.No) &&
								(choice.equalsIgnoreCase("n") || choice.equalsIgnoreCase("no") || choice.equalsIgnoreCase("(n)o"))) {
							return UserPrompt.Reply.No;

						} else if ((validReplies[i] == UserPrompt.Reply.Always) &&
								(choice.equalsIgnoreCase("a") || choice.equalsIgnoreCase("always") || choice.equalsIgnoreCase("(a)lways"))) {
							return UserPrompt.Reply.Always;

						} else if ((validReplies[i] == UserPrompt.Reply.Never) &&
								(choice.equalsIgnoreCase("r") || choice.equalsIgnoreCase("never") || choice.equalsIgnoreCase("neve(r)"))) {
							return UserPrompt.Reply.Never;
						}
					}
				}

			} catch (Exception e) {
				p("\nERROR: Cannot read from keyboard. Program aborted.\n");
				System.exit(42);
			}

			return null; // should never be reached
		}

	}
}

/*
BUILD NOTES:

Compile:
javac -Xlint ZipSnap.java

Package into JAR executable:
jar cvfm ZipSnap.jar manifest.txt ZipSnap*.class
*/
